/*
 * A quick and dirty polygon-based pathfinding implementation.
 * 
 * Credits:
 * - David Gouveia    | Detection of concave vertices and vertices within line-of-sight
 * - Antoine Vugliano | A* implementation
 * - KP               | A* node hierarchy creation from polygonal data, normal calculation, + fixes to Gouveia & Antoine's code
 */
#include "Pathfinder.h"

#include <math.h>

#include <algorithm>
#include <unordered_map>

namespace {

using namespace kp;

const float FUDGE = 0.05f; // Fudge factor for pushing OoB concave points back into the input polygon data

float Distance(XY p1, XY p2)
{
	return sqrtf((p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y));
}

bool CompareFloat(float a, float b, float epsilon = 0.000001f)
{
	return a > b - epsilon && a < b + epsilon;
}

std::vector<int> FindConcavePoints(const Polygon& polygon)
{
	std::vector<int> concave_points;
	size_t n = polygon.size();
	for (int i = 0; i < n; ++i)
	{
		XY a = polygon[(i - 1 + n) % n];
		XY b = polygon[(i)];
		XY c = polygon[(i + 1) % n];

		float cross = (b.x - a.x) * (c.y - b.y) - (b.y - a.y) * (c.x - b.x);
		if (cross >= 0)
			continue;

		concave_points.push_back(i);
	}
	return concave_points;
}

bool LineSegmentsCross(XY a, XY b, XY c, XY d)
{
	float denominator = ((b.x - a.x) * (d.y - c.y)) - ((b.y - a.y) * (d.x - c.x));

	if (CompareFloat(denominator, 0.0f))
		return false;

	float numerator1 = ((a.y - c.y) * (d.x - c.x)) - ((a.x - c.x) * (d.y - c.y));
	float numerator2 = ((a.y - c.y) * (b.x - a.x)) - ((a.x - c.x) * (b.y - a.y));

	if (CompareFloat(numerator1, 0.0f) || CompareFloat(numerator2, 0.0f))
		return false;

	float r = numerator1 / denominator;
	float s = numerator2 / denominator;

	return (r > 0 && r < 1) && (s > 0 && s < 1);
}

bool PointInPolygon(Polygon polygon, XY point, bool overlap_point_tolerance = true)
{
	const float epsilon = 0.5f;

	bool inside = false;

	if (polygon.size() < 3)
		return false;

	for (int i = 0, j = polygon.size() - 1; i < polygon.size(); j = i++)
	{
		float xi = polygon[i].x, yi = polygon[i].y;
		float xj = polygon[j].x, yj = polygon[j].y;

		if (overlap_point_tolerance)
		{
			if ((CompareFloat(point.x, xi) && CompareFloat(point.y, yi)) ||
				(CompareFloat(point.x, xj) && CompareFloat(point.y, yj)))
				return true;
		}

		// Perform a ray-cast to determine if the point is on the correct side of the edge
		if (((yi > point.y) != (yj > point.y)) && (point.x < (xj - xi) * (point.y - yi) / (yj - yi) + xi))
			inside ^= 1;
	}

	return inside;
}

bool PointInLine(float x1, float y1, float x2, float y2, float x0, float y0)
{
	return CompareFloat((y2 - y1) * x0 + (x1 - x2) * y0 + (x2 * y1 - x1 * y2), 0.0f);
}

bool InLineOfSight(XY start, XY end, const Polygon& polygon)
{
    // Start/end locations are the same -- consider these within LoS
    if (CompareFloat(Distance(start, end), 0.0f, 0.000001f))
        return true;

    // Start/end locations are outside the polygon -- don't consider these within LoS
    if (!PointInPolygon(polygon, start) || !PointInPolygon(polygon, end))
        return false;

    // Start/end segments intersect polygon edges -- should this be the case, don't consider these within LoS
    for (int i = 0; i < polygon.size(); i++)
    {
        XY line_start = polygon[i];
        XY line_end = polygon[(i + 1) % polygon.size()];
        if (LineSegmentsCross(start, end, line_start, line_end))
            return false;
    }

	// Center between start/end locations is outside the polygon and not part of a line segment -- should this be the case,
	// don't consider these within LoS. This addresses the same issue as the X/Y alignment hack above.
	XY middle = {
		(start.x + end.x) * 0.5f,
		(start.y + end.y) * 0.5f
	};
	bool in_middle = false;
	for (int i = 0; i < polygon.size(); i++)
	{
		XY line_start = polygon[i];
		XY line_end = polygon[(i + 1) % polygon.size()];
		if (PointInLine(middle.x, middle.y, line_start.x, line_start.y, line_end.x, line_end.y))
			in_middle = true;
	}
	if (!PointInPolygon(polygon, middle) && !in_middle)
		return false;

    // If no intersections and both points are inside, they are within LoS
    return true;
}

struct Node
{
	XY position{};
	float f = 0.0f, g = 0.0f, h = 0.0f; // A* stuff
	bool open = false, closed = false;  // A* stuff
	Node* parent = nullptr;
	std::vector<std::pair<Node*, float>> children;

	Node() {}
	Node(XY position): position(position) {}

	void AddChild(Node* child, float distance)
	{
		children.emplace_back(child, distance);
	}

	float DistanceTo(Node* other)
	{
		return Distance(position, other->position);
	}
};

// Comparator for STL heap and vector types
float CompareF(const Node* a, const Node* b)
{
	return a->f < b->f;
}

} // namespace

namespace kp {

bool FindPath(XY start, XY end, const Polygon& polygon, Path& path_out, PathMetadata* metadata)
{
	// Step 1 -- Build a node hierarchy.
	// We'll start by finding all of the concave points in the polygon; these will be our primary pathfinding nodes.
	std::vector<int> concaves = FindConcavePoints(polygon);
	std::vector<Node> nodes;
	std::vector<std::pair<size_t, size_t>> links;
	Node start_node(start);
	Node end_node(end);

	// Next calculate the normals of the polygon's vertices, allowing us to offset them to reduce the likelihood of encountering
	// weird edge cases in our line-of-sight detection. Doing this also lets us ensure that our generated path will be slightly
	// distant from the edges of the polygon, making it more suitable for something like an NPC in a game, which would otherwise
	// collide with the edges whilst traversing the path.
	std::unordered_map<int, XY> normals;
	std::unordered_map<int, XY> v_normals;
	for (size_t i = 0; i < polygon.size(); i++)
	{
		size_t j = (i + 1) % polygon.size();
		XY normal = {-(polygon[j].y - polygon[i].y), polygon[j].x - polygon[i].x};
		float normal_len = sqrtf(normal.x * normal.x + normal.y * normal.y);
		normal.x /= normal_len;
		normal.y /= normal_len;
		normals.emplace(i, normal);
		XY out1 = {polygon[i].x + (polygon[j].x - polygon[i].x) * 0.5f, polygon[i].y + (polygon[j].y - polygon[i].y) * 0.5f};
		XY out2 = {out1.x + normal.x, out1.y + normal.y};
		if (metadata)
			metadata->normal_lines.emplace_back(out1, out2);
	}

	for (const auto& [i, normal]: normals)
	{
		size_t j = (normals.size() + i - 1) % normals.size();
		XY v_normal = {normals[i].x + normals[j].x, normals[i].y + normals[j].y};
		float v_normal_len = sqrtf(v_normal.x * v_normal.x + v_normal.y * v_normal.y);
		v_normal.x /= v_normal_len;
		v_normal.y /= v_normal_len;
		v_normals.emplace(i, v_normal);
		if (metadata)
			metadata->normal_lines.emplace_back(polygon[i], XY{polygon[i].x + v_normal.x, polygon[i].y + v_normal.y});
	}

	// Now go through all of the concave points and create A* nodes from them. Additionally, link together ones which are in
	// line-of-sight of each other.
	for (size_t i = 0; i < concaves.size(); i++)
	{
		XY position = {polygon[concaves[i]].x + v_normals[concaves[i]].x, polygon[concaves[i]].y + v_normals[concaves[i]].y};

		// Should our vertex with the normal applied lands outside of the polygon, push it back in...
		while (!PointInPolygon(polygon, position))
		{
			position.x -= v_normals[concaves[i]].x * FUDGE;
			position.y -= v_normals[concaves[i]].y * FUDGE;
		}

		if (metadata)
			metadata->concave_points.push_back(position);
		nodes.emplace_back(position);

		for (size_t j = i + 1; j < concaves.size(); j++)
		{
			XY other_position = {
				polygon[concaves[j]].x + v_normals[concaves[j]].x, polygon[concaves[j]].y + v_normals[concaves[j]].y
			};

			while (!PointInPolygon(polygon, other_position))
			{
				other_position.x -= v_normals[concaves[j]].x * FUDGE;
				other_position.y -= v_normals[concaves[j]].y * FUDGE;
			}

			if (InLineOfSight(position, other_position, polygon))
			{
				if (metadata)
					metadata->linked_concave_points.emplace_back(position, other_position);
				links.emplace_back(i, j);
			}
		}
	}

	for (const auto& [i, j]: links)
	{
		nodes[i].AddChild(&nodes[j], nodes[i].DistanceTo(&nodes[j]));
		nodes[j].AddChild(&nodes[i], nodes[i].DistanceTo(&nodes[j]));
	}

	// Now figure out which nodes are in line-of-sight of our start and end positions.
	for (Node& node: nodes)
	{
		if (InLineOfSight(end_node.position, node.position, polygon))
		{
			node.AddChild(&end_node, node.DistanceTo(&end_node));
			if (metadata)
			{
				metadata->end_lines.push_back(end_node.position);
				metadata->end_lines.push_back(node.position);
			}
		}
		if (InLineOfSight(start_node.position, node.position, polygon))
		{
			start_node.AddChild(&node, node.DistanceTo(&start_node));
			if (metadata)
			{
				metadata->start_lines.push_back(start_node.position);
				metadata->start_lines.push_back(node.position);
			}
		}
	}

	// Step 2 -- We have our node hierarchy, now perform A* pathfinding.
	// This implementation is based on Antoine Vugliano's Pathfinder project.
	// Portions of this code are licensed under the BSD 2-clause license; see LICENSE.txt for details.
	float f = 0.0f, g = 0.0f, h = 0.0f;
	std::vector<Node*> open, closed;
	std::make_heap(open.begin(), open.end(), CompareF);
	open.push_back(&start_node);
	std::push_heap(open.begin(), open.end(), CompareF);
	start_node.open = true;
	Node* current_node = nullptr;
	while (!open.empty())
	{
		std::sort(open.begin(), open.end(), CompareF);

		current_node = open.front();
		std::pop_heap(open.begin(), open.end(), CompareF);
		open.pop_back();
		current_node->open = false;

		current_node->closed = true;
		closed.push_back(current_node);

		if (current_node == &end_node)
		{
			Node* parent = current_node->parent;
			path_out.push_back(current_node->position);
			while (parent != nullptr)
			{
				path_out.push_back(parent->position);
				parent = parent->parent;
			}
			return true;
		}

		for (const auto& [child, distance]: current_node->children)
		{
			g = current_node->g + distance;
			if ((child->open || child->closed) && child->g <= g)
				continue;

			h = Distance(child->position, end_node.position);
			f = g + h;
			child->f = f;
			child->g = g;
			child->h = h;
			child->parent = current_node;

			if (child->closed)
				child->closed = false;
			if (!child->open)
			{
				open.push_back(child);
				std::push_heap(open.begin(), open.end(), CompareF);
				child->open = true;
			}
		}
	}

	return false;
}

} // namespace kp
