#include <stdint.h>
#include <string.h>

#include <iostream>

#include <SDL2/SDL.h>

#include "Pathfinder.h"

using namespace kp;

namespace {

const float   SCALE = 18.0f;
SDL_Window*   window;
SDL_Renderer* renderer;
Polygon       polygon;
Path          path;
PathMetadata  path_metadata;
XY            path_start;
XY            path_end;
XY            mouse;
bool          show_debug = true;

void Init()
{
	path_start = {15, 6};
	path_end = {18, 22};
	polygon = {{3, 15}, {3, 3}, {24, 3}, {24, 8}, {6, 8}, {6, 10}, {24, 10}, {24, 24}, {15, 24}, {15, 17}, {21, 17}, {21, 15}};
}

void PerformPathfinding()
{
	path_metadata.concave_points.clear();
	path_metadata.linked_concave_points.clear();
	path_metadata.normal_lines.clear();
	path_metadata.start_lines.clear();
	path_metadata.end_lines.clear();
	path.clear();
	if (!FindPath(path_start, path_end, polygon, path, &path_metadata))
		std::cerr << "Failed to find path\n";
}

void DrawPoint(float x, float y, uint32_t color)
{
	SDL_FRect point_rect = { x * SCALE - SCALE * 0.25f, y * SCALE - SCALE * 0.25f, SCALE * 0.5f, SCALE * 0.5f };
	SDL_SetRenderDrawColor(renderer, (color >> 24) & 0xFF, (color >> 16) & 0xFF, (color >> 8) & 0xFF, (color) & 0xFF);
	SDL_RenderFillRectF(renderer, &point_rect);
}

void DrawLine(float x1, float y1, float x2, float y2, uint32_t color)
{
	SDL_SetRenderDrawColor(renderer, (color >> 24) & 0xFF, (color >> 16) & 0xFF, (color >> 8) & 0xFF, (color) & 0xFF);
	SDL_RenderDrawLineF(renderer, x1 * SCALE, y1 * SCALE, x2 * SCALE, y2 * SCALE);
}

void DrawPath(const Path& path, uint32_t color)
{
	if (path.empty())
		return;
	SDL_SetRenderDrawColor(renderer, (color >> 24) & 0xFF, (color >> 16) & 0xFF, (color >> 8) & 0xFF, (color) & 0xFF);
	for (int i = 0; i < path.size() - 1; i++)
	{
		XY start = path[i];
		XY end = path[i + 1];
		SDL_RenderDrawLineF(renderer, start.x * SCALE, start.y * SCALE, end.x * SCALE, end.y * SCALE);
	}
}

void DrawPolygon(const Polygon& polygon, uint32_t color)
{
	SDL_SetRenderDrawColor(renderer, (color >> 24) & 0xFF, (color >> 16) & 0xFF, (color >> 8) & 0xFF, (color) & 0xFF);
	for (int i = 0; i < path.size(); i++)
	{
		XY start = path[i];
		XY end = path[(i + 1) % path.size()];
		SDL_RenderDrawLineF(renderer, start.x * SCALE, start.y * SCALE, end.x * SCALE, end.y * SCALE);
	}
}

void DrawString(const char* txt, int x, int y)
{
	static uint64_t font[] = {0x7cfc3cfcfefe3cc6,0x7e7ec6c0c6c67cfc,0x7cfc7cfcc6c6c6c6,0xccfe187c0000600c,0x0000c6c666c6c0c0,
	       0x66c61818ccc0eee6,0xc6c6c6c6c630c6c6,0xc6c6cc0618c60000,0x30180000c6c6c0c6,0xc0c0c0c61818d8c0,0xfef6c6c6c6c6c030,
	       0xc6c6c66ccc0c1806,0x00001830fc00fefc,0xc0c6f8f8c0fe1818,0xf0c0d6dec6c6c6c6,0x7030c6c6c6387818,0x180c00000c600000,
		   0xc6c6c0c6c0c0cec6,0x1818d8c0c6cec6fc,0xc6fc1c30c6c6d66c,0x3030181800000c60,0x0000c6c6c0c6c0c0,0xc2c61818ccc0c6c6,
		   0xc6c0c6c60630c6c6,0xfec6306018300000,0x1830fc00c6c666c6,0xc0c066c61818c6c0,0xc6c6c6c0ccc6c630,0xc66ceec630c00000,
		   0x180830180000c6fc,0x3cfcfec03cc67e70,0xc6fec6c67cc07ac6,0x7c307c38c6c630fe,0x18301818600c0000,0x0c0000007c187c7c,
		   0xccfe7cfe7c7c7c30,0x467c6c000000d8c0,0xe0e060c030c03060,0x3c3c0c000000c638,0xc6c6ccc0c606c6c6,0xc678e6d66c00c0c0,
		   0xd8c0c060c0606060,0x30604242183000cc,0xc6180606ccc0c006,0xc6c6c6cc4cd0fe00,0xc0c04840c060c060,0x6060303099991830,
		   0x00ccd6180c1cfe7c,0xfc0c7c7e6c001870,0x6c6c00009080c060,0xc060c0303030a1a5,0x30fcfc30d6181806,0x0c06c618c6063000,
		   0x301c6cd800000000,0xc060c06060603018,0xa1a5303000ccc618,0x30060c06c630c606,0xd8006416fe00c040,0x0000c060c0606060,
		   0x3018999d603000cc,0xc61860c60cc6c630,0xc6c6cc00ced66c00,0xc0c00000c060c060,0x6060300c42466000,0x00007c3cfe7c0c7c,
		   0x7c307c7c7600c47c,0x6c0000000000e0e0,0x60c030c0300c3c3c};
	static const char*  font_mapping = "abcdefghijklmnopqrstuvwxyz!?.,><= /+-*0123456789&^%$#~:;\"'[](){}|\\`@";
	static SDL_Surface* font_surface = NULL;
	static SDL_Texture* font_texture = NULL;
	if (!font_surface)
	{
		font_surface = SDL_CreateRGBSurfaceWithFormat(0, 272, 16, 32, SDL_PIXELFORMAT_ABGR8888);
		for (int i = 0; i < 272 * 16; i++)
			((uint32_t*)font_surface->pixels)[i] = ((font[i >> 6] >> (63 - (i & 63))) & 1) * 0xFFFFFFFF;
		font_texture = SDL_CreateTextureFromSurface(renderer, font_surface);
	}
	for (int i = 0; i < strlen(txt); i++)
	{
		const char* pos = strchr(font_mapping, tolower(txt[i]));
		size_t ci = pos ? pos - font_mapping : 0;
		SDL_Rect src_rect = {(ci % 34) * 8, (ci / 34) * 8, 8, 8};
		SDL_Rect dst_rect = {x + i * 8, y, 8, 8};
		SDL_RenderCopy(renderer, font_texture, &src_rect, &dst_rect);
	}
}

void Render()
{
	// Draw map
	SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND);
	SDL_SetRenderDrawColor(renderer, 0xFF, 0xFF, 0xFF, 0xFF);
	for (int i = 0; i < polygon.size(); i++)
	{
		XY start = polygon[i];
		XY end = polygon[(i + 1) % polygon.size()];
		SDL_RenderDrawLineF(renderer, start.x * SCALE, start.y * SCALE, end.x * SCALE, end.y * SCALE);
	}

	// Draw start/end points
	DrawPoint(path_start.x, path_start.y, 0xFF0000FF);
	DrawPoint(path_end.x, path_end.y, 0x0000FFFF);

	// Draw path metadata
	if (show_debug)
	{
		// Draw data gathered from the pathfinding process
		for (XY xy : path_metadata.concave_points)
			DrawPoint(xy.x, xy.y, 0x00FFFFFF);
		for (const auto& [i, j]: path_metadata.linked_concave_points)
			DrawLine(i.x+1/SCALE, i.y+1/SCALE, j.x+1/SCALE, j.y+1/SCALE, 0xFF6600FF); // Offset a tiny bit for visibility
		for (const auto& [i, j]: path_metadata.normal_lines)
			DrawLine(i.x, i.y, j.x, j.y, 0xFF00FFFF);
		DrawPath(path_metadata.start_lines, 0xFF00007F);
		DrawPath(path_metadata.end_lines, 0x0000FF7F);

		// Draw preview
		DrawLine(polygon.back().x, polygon.back().y, mouse.x / SCALE, mouse.y / SCALE, 0xFFFFFF55);
	}

	// Draw generated path
	DrawPath(path, 0x00AA00FF);

	// """GUI"""
	DrawString("Polygon-based A* implementation", 0, 0);
	DrawString("-------------------------------", 0, 9);
	DrawString("MOUSE1 = Draw Polygons, MOUSE2 = Clear Screen, S = move start, E = move end", 0, 18);
	DrawString("X = Find Path, LSHIFT = Toggle path metadata & polygon edit preview", 0, 27);
}

} // namespace

/*
 * SDL boilerplate
 */
int main(int, char*[])
{
	SDL_Init(SDL_INIT_EVERYTHING);
	window = SDL_CreateWindow("Pathfinding", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 640, 480, SDL_WINDOW_SHOWN);
	renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_PRESENTVSYNC);

	Init();

	for (;;)
	{
		for (SDL_Event e; SDL_PollEvent(&e); )
		{
			switch (e.type)
			{
			case SDL_QUIT:
				goto cleanup;
			case SDL_MOUSEMOTION:
				mouse.x = (float)e.motion.x;
				mouse.y = (float)e.motion.y;
				break;
			case SDL_KEYDOWN:
				switch (e.key.keysym.scancode)
				{
				case SDL_SCANCODE_X:
					PerformPathfinding();
					break;
				case SDL_SCANCODE_S:
					path_start.x = mouse.x / SCALE;
					path_start.y = mouse.y / SCALE;
					break;
				case SDL_SCANCODE_E:
					path_end.x = mouse.x / SCALE;
					path_end.y = mouse.y / SCALE;
					break;
				case SDL_SCANCODE_LSHIFT:
					show_debug ^= 1;
					break;
				}
				break;
			case SDL_KEYUP:
				break;
			case SDL_MOUSEBUTTONDOWN:
				switch (e.button.button)
				{
				case 1:
					polygon.push_back({mouse.x / SCALE, mouse.y / SCALE});
					break;
				case 2:
					break;
				case 3:
					polygon.clear();
					break;
				}
				break;
			}
		}

		SDL_SetRenderDrawColor(renderer, 0x00, 0x00, 0x55, 0x00);
		SDL_RenderClear(renderer);
		Render();
		SDL_RenderPresent(renderer);
	}

cleanup:
	SDL_DestroyRenderer(renderer);
	SDL_DestroyWindow(window);
	SDL_Quit();

	return 0;
}
